unit DJTSatellites;

{
Revision history:

1.0  1999 Jul 12  First version         
1.1  1999 Aug 28  Add VersionInfoString property
                  Return pass times with seconds
                  Require minimum pass intervals in seconds
1.2  1999 Sep 25  Faster, more accurate algorithm
                  Return more pass information, include seconds
1.3  1999 Sep 25  More complete pass info - add years
1.4  1999 Sep 30  Return actual version string
2.0  2000 Feb 05  Rebuild in simplified form
3.0  2001 Jan 10  Version built with Delphi 5..
                  .. but avoid requiring Delphi 5 runtime library
                  Pass list now includes 4-digit year
                  Add 'SUN' and 'MOON' as known satellite names
                  Ensure parameters ordered <lon> <lat>
                  Return Range and RangeRate and Phase
                  .. Southbound & RangeRate not for sun and moon
                  .. Phase (for moon only)
3.1  2001 May 17  Expose RevNumber and OrbitPhase
                  Add Helmut Strickner's FindAOS and FindLOS functions
                  Add MeanAnomaly as real number of orbits
3.2  2001 Nov 17  Make satellite state readable
3.3  2002 Aug 24  Make Period and Kepler data accessible
3.4  2004 Jul 21  Add MakePassDetails method
                  Rearrange routines alphabetically
3.5  2005 Jul 17  Add IgnoreRefraction property
                  Allow for KeplerPath to be a file as well as a directory
                  Fix memory leak with duplicate satellite names
                  Increase error-checking in MakePassDetails ..
                  .. where a direction other than N, n, S, s is specified
                  .. where the satellite name is mis-spelt
3.6  2006 Sep 02  Add MakePassDetails2 method, adds range rate
}
                                          
interface

uses
  Windows, ComObj, ActiveX, DJTSatLib_TLB,
  Satellite, SGP_Math, SGP_Constants, SGP_Time, Dialogs;

type
  TSatellites = class (TAutoObject, ISatellites)
  private
    FSatelliteList: TSatelliteList;
    FDataPath: string;
    Fobs_gd: vector;
    FRevNumber: integer;
    FOrbitPhase: double;
    FMeanAnomaly: double;
    Fsat_pos, Fsat_vel: vector;
  protected
    { Protected declarations }
    function Get_VersionString: WideString; safecall;
    function Get_KeplerPath: WideString; safecall;
    procedure Set_KeplerPath (const Value: WideString); safecall;
    function Get_SatCount: Integer; safecall;
    function Get_SatNames: WideString; safecall;
    function Get_ObsLat: Double; safecall;
    procedure Set_ObsLat (Value: Double); safecall;
    function Get_ObsLon: Double; safecall;
    procedure Set_ObsLon (Value: Double); safecall;
    function Get_ObsHeight: Double; safecall;
    procedure Set_ObsHeight (Value: Double); safecall;
    function PassListForPeriod (const Name: WideString; UTCDate: Double;
      MinSeconds, NumDays: Integer; MinElevDegrees: Double): WideString;
      safecall;
    function IsVisible (const Name: WideString; JulianTime,
      ThresholdDeg: Double; out Southbound: WordBool; out Azimuth,
      Elevation, Longitude, Latitude,
      Range, RangeRate, Altitude,
      Phase: Double): WordBool; safecall;
    function Get_RevNumber: Integer; safecall;
    function Get_OrbitPhase: Double; safecall;
    function Get_MeanAnomaly: Double; safecall;
    function DateTimeToJT (DateTime: double): double; safecall;
    function FindAOS (const Name: WideString; PredictTime: double): double; safecall;
    function FindLOS (const Name: WideString; PredictTime: double): double; safecall;
    procedure SatState (out X, Y, Z, Xdot, Ydot, Zdot: Double); safecall;
    function Period (const Name: WideString): double; safecall;
    procedure KeplerData (const Name: WideString;
                          out JulianEpoch: Double;
                          out Inclination: Double;
                          out RAAN: Double;
                          out Eccentricity: Double;
                          out ArgOfPerigee: Double;
                          out DecayRate: Double;
                          out Period: Double;
                          out MeanAnomaly: Double) safecall;
    function  MakePassDetails (const PassDescription: WideString; StepSeconds: Integer): WideString; safecall;
    procedure Initialize; override;
    function  Get_IgnoreRefraction: WordBool; safecall;
    procedure Set_IgnoreRefraction (Value: WordBool); safecall;
    function  MakePassDetails2 (const PassDescription: WideString; StepSeconds: Integer): WideString; safecall;
  end;


implementation

uses SysUtils, ComServ, VectorOrbit;

var
  VersionInfoString: string;

var
  dll_path: array [0..255] of char;
  dummy: DWord;
  FVersionInfoSize: integer;
  FVersionInfo: PChar;
  FixedInfoData: PVSFixedFileInfo;
  QueryLen: UINT;


function TSatellites.DateTimeToJT (DateTime: double): double;
begin
  Result := Satellite.DateTimeToJT (DateTime);
end;


// This code from: "Helmut Strickner" <hstr@gil.com.au>

const
  FindStep: double = (1/1440)*15; // (minutes) initial step size in finding aos/los
  FindAngleError: double = 0.05; //  (degrees) if within this angle quit
  FindStepMinimum: double = (1/1440)/10; // (minutes)  minimum step size
  ObsAngle: double = 0.5; //Sat elevation considered as 'UP'

{ if Result= -1 then AOS/LOS is more than two days away }
function TSatellites.FindAOS (const Name: WideString;
                              PredictTime: double): double;
var
  step_time, totelapse_time, neg_time, time_now: TDateTime;
  neg_el: double;
  jt: double;
  southbound: WordBool;
  az, el, lon, lat: double;
  range, range_rate, altitude, phase: double;
begin
  neg_time := 0;
  time_now := PredictTime;
  jt := DateTimeToJT (time_now);
  IsVisible (Name, jt, ObsAngle, southbound, az, el, lon, lat,
                   range, range_rate, altitude, phase);

{  if satellite is currently up (Elevation > user value)
   then AOS problem solved (e.g. 0.0)...find LOS.  }

  if (el >= ObsAngle)
  then  // later horizon profile from file - ObsAngle(az))
    begin
    Result := time_now;
    Exit;
    end
  else
    begin

{   StepSize algorithm based on work by NZ3F and
    WN2A (author of SatSked)

    use current elevation angle to determine stepsize
    big angle....big stepsize  }


    step_time := FindStep*ABS (el*el);

{  step time forward and determine elevation angle.
   when we find an elevation angle > requirement then
   we linear interpolate between last negative answer
   and this one to get the answer }

    totelapse_time := 0;
    while (el < 0) do  // ObsAngle
      begin
      neg_el := el;
      neg_time := time_now;
      time_now := time_now + step_time;
      totelapse_time := totelapse_time + step_time;
      if (totelapse_time > 2.0) then
        begin
        Result := -1; //?? more than two days
        Exit;
        end;
      jt := DateTimeToJT (time_now);
      IsVisible (Name, jt, ObsAngle, southbound, az, el, lon, lat,
                       range, range_rate, altitude, phase);
      step_time := FindStep * ABS (el*el);
      if (step_time < FindStepMinimum) then step_time := FindStepMinimum;
    end;
// if within error then return  - this gives an error of up to a few minutes
 {   if (ABS(el) < FindAngleError)then  // so I always use the 1/10 step size as well
    begin
      Result:= time_now;
      Exit;
    end;    }
// get better answer - start over with 1/10 step size
    time_now := neg_time; //start off with previous time value
    el := neg_el;       //and previous el value
    step_time := step_time / 10.0;
    while (el < 0) do // ObsAngle
      begin
      neg_el := el;
      time_now := time_now + step_time;
      jt := DateTimeToJT (time_now);
      IsVisible (Name, jt, ObsAngle, southbound, az, el, lon, lat,
                       range, range_rate, altitude, phase);
      step_time := FindStep * ABS (el*el);
      if (step_time < FindStepMinimum) then step_time:= FindStepMinimum;
      end;
    Result := time_now;
  end;
end;


function TSatellites.FindLOS (const Name: WideString;
                              PredictTime: double): double;
var
  step_time, totelapse_time, pos_time, time_now: TDateTime;
  pos_el: double;
  jt: double;
  southbound: WordBool;
  az, el, lon, lat: double;
  range, range_rate, altitude, phase: double;
begin
  pos_time:= 0;
  time_now := PredictTime;
  jt := DateTimeToJT (time_now);
  IsVisible (Name, jt, ObsAngle, southbound, az, el, lon, lat,
                   range, range_rate, altitude, phase);

 { if (el >= ObsAngle)then  // later horizon profile from file - ObsAngle(az))
  begin
    Result:= time_now;
    Exit;
  end else }
  begin
    step_time := FindStep * ABS (el*el);

    totelapse_time := 0;
    while (el > 0) do  // ObsAngle
      begin
      pos_el := el;
      pos_time := time_now;
      time_now := time_now + step_time;
      totelapse_time := totelapse_time + step_time;
      if (totelapse_time > 2.0) then
        begin
        Result:= -1; //?? more than two days
        Exit;
        end;
      jt := DateTimeToJT (time_now);
      IsVisible (Name, jt, ObsAngle, southbound, az, el, lon, lat,
                       range, range_rate, altitude, phase);
      step_time := FindStep * ABS (el*el);
      if (step_time < FindStepMinimum) then step_time := FindStepMinimum;
    end;
// if within error then return
 {   if (ABS(el) < FindAngleError)then
    begin
      Result:= time_now;
      Exit;
    end;    }
// get better answer - start over with 1/10 step size
    time_now := pos_time; //start off with previous time value
    el := pos_el;       //and previous el value
    step_time := step_time / 10.0;
    while (el > 0) do // ObsAngle
      begin
      pos_el := el;
      time_now := time_now + step_time;
      jt := DateTimeToJT (time_now);
      IsVisible (Name, jt, ObsAngle, southbound, az, el, lon, lat,
                       range, range_rate, altitude, phase);
      step_time := FindStep * ABS (el*el);
      if (step_time < FindStepMinimum) then step_time := FindStepMinimum;
      end;
    Result := time_now;
  end;
end;


function TSatellites.Get_IgnoreRefraction: WordBool;
begin
  Result := SGP_Constants.IgnoreRefraction;
end;


function TSatellites.Get_KeplerPath: WideString;
begin
  Result := FDataPath;
end;


function TSatellites.Get_MeanAnomaly: Double;
begin
  Result := FMeanAnomaly;
end;


function TSatellites.Get_ObsLat: Double;
begin
  Result := Degrees (Fobs_gd [1]);
end;


function TSatellites.Get_ObsLon: Double;
begin
  Result := Degrees (Fobs_gd [2]);
end;


function TSatellites.Get_ObsHeight: Double;
begin
  // Return value in metres (internally, km)
  Result := 1000.0 * Fobs_gd [3];
end;


function TSatellites.Get_OrbitPhase: Double;
begin
  Result := FOrbitPhase;
end;


function TSatellites.Get_RevNumber: Integer;
begin
  Result := FRevNumber;
end;


function TSatellites.Get_SatCount: Integer;
begin
  Result := FSatelliteList.Count;
end;


function TSatellites.Get_SatNames: WideString;
begin
  Result := FSatelliteList.CommaText;
end;


function TSatellites.Get_VersionString: WideString;
begin
  Result := VersionInfoString;
end;


procedure TSatellites.Initialize;
begin
  Inherited;
  // Create the list at this time so that there is a valid count
  FSatelliteList := TSatelliteList.Create ('');
  Set_ObsLat (56);
end;


function TSatellites.IsVisible (const Name: WideString; JulianTime,
  ThresholdDeg: Double;  out Southbound: WordBool;  out Azimuth, Elevation,
  Longitude, Latitude, Range, RangeRate, Altitude, Phase: Double): WordBool;
const
  Rad2Deg = 180 / pi;
var
  threshold: double;
  sat_num: integer;
  sat: TSatellite;
  sat_gd: vector;
  obs: vector;
  visible: boolean;
  known: boolean;
  uc_name: string;
  jt0: double;
  t: double;
begin
  Result := False;
  Phase := 0.0;
  RangeRate := 0.0;
  FRevNumber := 0;
  FOrbitPhase := 0.0;
  FMeanAnomaly := 0.0;
  // Degrees required above horizon, convert to radians
  threshold := ThresholdDeg * 2.0 * pi / 360.0;
  known := False;
  if FSatelliteList.Find (Name, sat_num)
  then
    begin
    known := True;
    sat := TSatellite (FSatelliteList.Objects [sat_num]);
    sat.SGP (JulianTime, Fsat_pos, Fsat_vel);
    FRevNumber := sat.Revolution;
    FOrbitPhase := sat.OrbitAngle;
    FMeanAnomaly := sat.MeanAnomaly;
    Convert_Sat_State (Fsat_pos, Fsat_vel);
    Calculate_LatLonAlt (Fsat_pos, JulianTime, sat_gd);
    end
  else
    begin
    Fsat_vel [1] := 0;
    Fsat_vel [2] := 0;
    Fsat_vel [3] := 0;
    Fsat_vel [4] := 0;
    jt0 := Julian_Date_of_Year (1899) + DOY (1899, 12, 30);
    t := JulianTime - jt0;
//    ShowMessage (FormatDateTime ('yyyy-mm-dd hh:nn:ss', t));
    uc_name := UpperCase (Name);
    if uc_name = 'SUN' then
      begin
      GetSolarPos (t, Fsat_pos, sat_gd);
      known := True;
      end;
    if uc_name = 'MOON' then
      begin
      GetLunarPos (t, Fsat_pos, sat_gd, phase);
      known := True;
      end;
    end;

  if known then
    begin
    Calculate_Obs (Fsat_pos, Fsat_vel, Fobs_gd, JulianTime, obs, visible);
    Azimuth := obs [1];
    Elevation := obs [2];
    Range := obs [3];
    RangeRate := obs [4];
    Result := obs [2] > threshold;
    Southbound := Fsat_vel [3] < 0;
    Latitude := sat_gd [1];
    Longitude := sat_gd [2];
    Altitude := sat_gd [3];
//    ShowMessage (Format ('Lon: %f, Lat: %f', [Rad2Deg * Longitude, Rad2Deg * Latitude]));
    end;
end;


procedure TSatellites.KeplerData (const Name: WideString;
  out JulianEpoch: Double;
  out Inclination: Double;
  out RAAN: Double;
  out Eccentricity: Double;
  out ArgOfPerigee: Double;
  out DecayRate: Double;
  out Period: Double;
  out MeanAnomaly: Double);
var
  sat_num: integer;
  sat: TSatellite;
begin
  if FSatelliteList.Find (Name, sat_num)
  then
    begin
    sat := TSatellite (FSatelliteList.Objects [sat_num]);
    JulianEpoch := sat.JEpoch;
    Inclination := sat.Inclination;
    RAAN := sat.RAAN;
    Eccentricity := sat.Eccentricity;
    ArgOfPerigee := sat.ArgOfPerigee;
    DecayRate := sat.DecayRate;
    Period := sat.Period;
    MeanAnomaly := sat.MeanAnomalyAtEpoch;  // radians
    end;
end;


function TSatellites.MakePassDetails (const PassDescription: WideString;
                                      StepSeconds: Integer): WideString;
begin
  Result := FSatelliteList.MakePassDetails (PassDescription, FObs_gd, StepSeconds);
end;


function TSatellites.MakePassDetails2 (const PassDescription: WideString;
  StepSeconds: Integer): WideString;
const
  IncludeRangeRate = True;
begin
  Result := FSatelliteList.MakePassDetails (PassDescription, FObs_gd, StepSeconds, IncludeRangeRate);
end;


function TSatellites.PassListForPeriod (const Name: WideString; UTCDate: Double;
  MinSeconds, NumDays: Integer; MinElevDegrees: Double): WideString;
var
  sat_num: integer;
  sat: TSatellite;
begin
  Result := '';
  if FSatelliteList.Find (Name, sat_num) then
    begin
    sat := TSatellite (FSatelliteList.Objects [sat_num]);
    Result := sat.PassListForPeriod
              (UTCDate, MinSeconds, NumDays, MinElevDegrees, Fobs_gd);
    end;
end;


function TSatellites.Period (const Name: WideString): double;
var
  sat_num: integer;
begin
  Result := 0.0;
  if FSatelliteList.Find (Name, sat_num)
  then
    begin
    Result := TSatellite (FSatelliteList.Objects [sat_num]).Period;
    end;
end;

procedure TSatellites.SatState (out X, Y, Z, Xdot, Ydot, Zdot: Double);
begin
  X := Fsat_pos [1];
  Y := Fsat_pos [2];
  Z := Fsat_pos [3];
  Xdot := Fsat_vel [1];
  Ydot := Fsat_vel [2];
  Zdot := Fsat_vel [3];
end;


procedure TSatellites.Set_IgnoreRefraction (Value: WordBool);
begin
  SGP_Constants.IgnoreRefraction := Value;
end;


procedure TSatellites.Set_KeplerPath (const Value: WideString);
begin
  FDataPath := Value;
  if FileExists (FDataPath) then FDataPath := ExtractFilePath (FDataPath);
  if Assigned (FSatelliteList) then
    begin
    FSatelliteList.Free;
    FSatelliteList := nil;
    end;
  if Value <> '' then FSatelliteList := TSatelliteList.Create (FDataPath);
end;


procedure TSatellites.Set_ObsHeight (Value: Double);
begin
  // Internally, store in km.  User interface is in metres.
  Fobs_gd [3] := 0.001 * Value;
end;


procedure TSatellites.Set_ObsLat (Value: Double);
begin
  Fobs_gd [1] := Radians (Value);
end;


procedure TSatellites.Set_ObsLon (Value: Double);
begin
  Fobs_gd [2] := Radians (Modulus (Value, 360));
end;


initialization
  TAutoObjectFactory.Create(ComServer, TSatellites, Class_Satellites,
    ciMultiInstance, tmApartment);

  // This code simply to get the version info string
  // directly from the DLL

  GetModuleFileName (HInstance, @dll_path[0], 255);
  FVersionInfoSize := GetFileVersionInfoSize (@dll_path[0], dummy);
  if FVersionInfoSize = 0
  then
  else
    begin
    GetMem (FVersionInfo, FVersionInfoSize);
    GetFileVersionInfo (@dll_path [0], dummy, FVersionInfoSize, FVersionInfo);
    VerQueryValue (FVersionInfo, '\', pointer(FixedInfoData), QueryLen);
    with FixedInfoData^ do
      VersionInfoString := Format ('%d.%d.%d.%d',
          [dwFileVersionMS shr 16, dwFileVersionMS and $FFFF,
           dwFileVersionLS shr 16, dwFileVersionLS and $FFFF]);
    end;

finalization
  if Assigned (FVersionInfo) then FreeMem (FVersionInfo);


end.

